#!/usr/bin/perl 
use strict;

###########################################################################
## function for skipping space characters
sub skipSpace {
    my $array_ref = $_[0];
    my $start = $_[1];
    while ($array_ref->[$start] =~ /\s/) { $start += 1; }
    return $start;
}

###############################################################################
## function for beautifying ugly LaTeX strings

sub beautify {
    my $hash = "gf654ho564h6u54i56dy456456t4654x56cfv456xb456cbn46g5h";
    my $str = $_[0];
# remove the bordering braces
    $str =~ s{ ^\{ \s* }{}xg;
    $str =~ s{ \s*\}$ }{}xg;

    $str =~ s{ \\ \{ }{$hash}xg;
    $str =~ s{ \{ }{}xg;
    $str =~ s{ $hash }{\{}xg;

    $str =~ s{ \\ \} }{$hash}xg;
    $str =~ s{ \} }{}xg;
    $str =~ s{ $hash }{\}}xg;

#    $str =~ s{ \\ \} }{$1}xg;
   
#    $str =~ s{ \\ \{ }{\{}xg;
#    $str =~ s{ \\ \} }{\}}xg;
    $str =~ s{ \\ \` u }{&\#249;}xg;
    $str =~ s{ \\ \' u }{&\#250;}xg;
    $str =~ s{ \\ \" u }{&\#252;}xg;
    $str =~ s{ \\ \` a }{&\#224;}xg;
    $str =~ s{ \\ \' a }{&\#225;}xg;
    $str =~ s{ \\ \" a }{&\#228;}xg;
    $str =~ s{ \\ \` e }{&\#232;}xg;
    $str =~ s{ \\ \' e }{&\#233;}xg;
    $str =~ s{ \\ \" e }{&\#235;}xg;
    $str =~ s{ \\ \` o }{&\#242;}xg;
    $str =~ s{ \\ \' o }{&\#243;}xg;
    $str =~ s{ \\ \" o }{&\#246;}xg;
    $str =~ s{ \\ \" \\i }{&\#239;}xg;
    $str =~ s{ \\ \' \\i }{&\#237;}xg;
    $str =~ s{ \\ \` \\i }{&\#236;}xg;
    return $str;
}


###############################################################################
## function for making a string out of a character array
sub makeString {
    my $array_ref = $_[0];
    my $start = $_[1];
    my $end = $_[2];

    my $string = join "",@{$array_ref}[$start .. $end - 1];

    return $string;
}

###############################################################################
## function for getting the boundaries of a string(inside or not inside a brace)
sub getStringEnds {

    my $array_ref = $_[0];
    my $start_loc = $_[1];
    my $end_loc = $#{$array_ref};
  
    my $done = 0;
    my $loc = $start_loc;
    my $level = 0;
    my $stringStart = -1;
    my $char = $array_ref->[4];
   
    
    while(!$done) {
	$char = $array_ref->[$loc];
	if ($char eq "\\") { 
	    $loc += 1;
	}
	else {
	    if ($char =~ /\w/ && $stringStart == -1) { $stringStart = $loc; }
	    if ($char eq '{') { 
		$level += 1; 
		if($stringStart == -1) {$stringStart = $loc;} 
	    }
	    if ($char eq '}') { $level -= 1; if ($level == 0) {$loc += 1; last;} }
	    if ($char eq "," && $level == 0) { 
		
		last; 
	    }
	    if ($char =~ /\s/ && $level == 0 && $array_ref->[$stringStart] ne '{') { 
		
		last; 
	    }
	    if ($char eq "=" && $level == 0) { 
		
		last; 
	    }
	}
	$loc += 1;
	if ($end_loc < $loc) { $done = 1; }
    }
    return $stringStart, $loc;
}

###############################################################################
## function for processing a row of a bib entry till a ',' at the existing level. Takes an array, string and location.

sub extractRow {
    my $array_ref = $_[0];
    my $start_loc = $_[1];
    my $end_loc = $#{$array_ref};
    my $done = 0;
    my $loc = $start_loc;
    my $char;

    my $typeStart;
    my $typeEnd;
    my $stringStart;
    my $stringEnd;

    my $type;
    my $value;

# get the lhs

    $loc = skipSpace($array_ref,$loc);
    if($array_ref->[$loc] eq '}') { return "message", "End of Bib Entry", $start_loc; }
   
    @_ = getStringEnds($array_ref,$loc);
    $type = makeString($array_ref,$_[0], $_[1]);

# pass the '='

    $loc = skipSpace($array_ref,$_[1]);
    
    if($array_ref->[$loc] eq ',') { 
	$value = $type; $type = "id";
	return $type, $value, $loc + 1;
    }
    
    if($array_ref->[$loc] eq '=') {
	$loc += 1;
	
	$loc = skipSpace($array_ref,$loc);
	
	

	#get the rhs
	@_ = getStringEnds($array_ref,$loc);
	
	$value = makeString($array_ref,$_[0], $_[1]);
	$loc = $_[1];
	while($array_ref->[$loc-1] ne "," && $array_ref->[$loc] ne "}") { $loc += 1; }
	
	return $type, $value, $loc; 
    }

    die "Parsing error at ^" . makeString($array_ref,$loc, $loc + 20);

}


###############################################################################
## function for processing authors
sub getAuthorSortKey {
    my $key = "";
    my $ath = "";
    my $a = $_[0];
    my @al = split / and /, $a;
    for my $i (0 .. $#al) {
	if($i != 0) {
	    $ath = $ath . ", ";
	}
	$ath = $ath . $al[$i];
	$al[$i] = [ split /\s+/, $al[$i] ];
	
	$key = $key . $al[$i]->[$#{$al[$i]}];
    }
    
    return $key, $ath;
}

###############################################################################
## function for processing an entry. takes a string and the starting location, returns index and ending location.



sub extractEntry {
    my $array_ref = $_[0];
    my $start_loc = $_[1];
    my $db_ref = $_[2];
    my $num = $_[3];
    my $str_ref = $_[4];
    my $end_loc = $#{$array_ref};
    my $done = 0;
    my $loc = $start_loc;
    my $char;

    my $typeStart;
    my $typeEnd;
    my $stringStart;
    my $stringEnd;

    my $state = 0; 

# find the @
    while($loc <= $end_loc && $array_ref->[$loc] ne '@') { $loc += 1; }
    
    if($loc > $end_loc) { return "", $loc; }

#get the word
    $loc += 1;
    $typeStart = $loc;
    while($array_ref->[$loc] =~ /\w/) { $loc += 1; }
    my $type = makeString($array_ref, $typeStart, $loc);
    $db_ref->[$num]{"type"} = lc($type);
# call string boundary
    $loc = skipSpace($array_ref, $loc);
    @_ = getStringEnds($array_ref, $loc);

    $loc += 1;
    
#in a while loop, process every entry.
    while (1) {
	@_ = extractRow($array_ref, $loc);
	$_[0] = lc($_[0]);
	if($loc == $_[2]) { last; }
	if(lc($type) eq "string" ) 
	{ 
	    $str_ref->[0]{$_[0]} = beautify($_[1]); 
	}
	else { 
	    $db_ref->[$num]{$_[0]} = beautify($_[1]); 
	}
	if($_[0] eq "author") {
	    my @aths = getAuthorSortKey($db_ref->[$num]{$_[0]});
	    $db_ref->[$num]{"sortkey"} = $aths[0];
	    $db_ref->[$num]{"author"} = $aths[1];
	}
	
	$loc = $_[2];
    }
    
    $loc = skipSpace($array_ref, $loc);
    $loc += 1;

    return $type, $loc;

}

# split at the first `{' after the first character. The first section is the type of the bib entry.

# call row processing function till string is over.



###############################################################################
## function to compare
sub itemCompare{ 
   
    return $_[0]->{"sortkey"};
   
}

###############################################################################
## function to sort

sub mySort {

    my $array_ref = $_[0];
   
    return [ sort {itemCompare($a) cmp itemCompare($b)} @{$array_ref} ];
}

###############################################################################
## main section

# read the input file as it is

# first, process the input file and remove the comments that start with '###'

# in a loop till the string is over, substring between an @ and } (included) and pass it onto a procedure to handle it.

# exit


open(BIB,"<$ARGV[0]");
my @file = <BIB>;
close(BIB);
my $a = "";

for my $i (@file) {
    $a = $a . $i;
}

$a =~ s{%%%.*\n}{<cl>}xg;

#print "\nafter comment replacement = " . $a;



my @array = split /^?/, $a;
my $arr_ref = [@array];
my $end_loc = $#{$arr_ref};


my $db_ref = [ ( ) ];
my $num = 0;

my $str_ref = [ ( ) ];


$_[1] = 0;

while(1) {
    @_ = extractEntry($arr_ref, $_[1], $db_ref, $num, $str_ref);
    
    if($_[1] > $end_loc) { last; }
    if($_[0] ne "string") {$num += 1;}
}

#for my $row ( keys %{ $str_ref->[0] } ) {
#    print "\n\t" . "$row => " . $str_ref->[0]{$row};
#}

for my $i ( 0 .. $#{$db_ref}) {
   
    for my $row ( keys %{ $db_ref->[$i] } ) {
	
	if(exists($str_ref->[0]{ $db_ref->[$i]{$row} })) {
	    
	    $db_ref->[$i]{$row} = $str_ref->[0]{ $db_ref->[$i]{$row} };
	}
	
    }
}


$db_ref = mySort($db_ref);

#for my $i ( 0 .. $#{$db_ref}) {
#   print "\n\nBib Entry No." . ($i + 1);
#    for my $row ( keys %{ $db_ref->[$i] } ) {
#	print "\n\t" . "$row => " . $db_ref->[$i]{$row};
#    }
#}

print "\n";
print "\n";




# PARSING IS OVER - BEGIN STORING THE HTML FILE####################

my %month = ("jan", "January", "feb", "February", "mar", "March", "apr", "April", "may", "May", "jun", "June", "jul", "July", "aug", "August", "sep", "September", "oct", "October", "nov", "November", "dec", "December");


# define a basic header
sub HtmlHeader
{
my $fileName = $ARGV[1];
unless(open(OUTFILE, ">$fileName"))
	{ die ("Couldn't open the output file\n"); }

print OUTFILE "\<HTML\>\n<HEAD>\n\t<TITLE>\n\tBibliography\n\t<\/TITLE>\n\n\<STYLE TYPE=\"TEXT/CSS\"\>table { width : 600px; } \n\ntd { text-align : justify; vertical-align: top;\n}\<\/STYLE\><\/HEAD><BODY>\n\n";
print OUTFILE "\<H1>References:\n\n<\/H1>\n";
}

# define an end of the html file
sub HtmlTail
{

print OUTFILE ("\n\<\/BODY>\n\<\/HTML>\n");
}

my $key;
&HtmlHeader;
&AddToHtml;
&HtmlTail;






# add  all entries to the html file
sub AddToHtml
{
my $key;
my $id;
my $db_count = $#{ $db_ref } + 1;

print OUTFILE "<TABLE>\n";
for(my $i = 0; $i < $db_count; $i++)
	{
	my %entry = %{$db_ref->[$i]};

	# IF ARTICLE############################################################
	# REQUIRED FIELDS - IF MISSING GENETRATE WARNING - AUTHOR TITLE JOUNAL YEAR
	# OPTIONAL FIELDS - IF PRESENT DISPLAY - VOL, NO, PAGES, MONTH
	####################################################################

	if(lc($entry{"type"}) eq lc("Article"))
	{

	# remove the AND and replace them with commas to seperate all authors and last one by "and"
	# AUTHOR ADDITION 
	if(!exists( $entry{"author"}) )
		{ print "Warning: empty Author in Article\n";}
	else
		{
		my $authors = $entry{"author"};
		my @authorList = split(/ and /, $authors);
		my $count = @authorList."\n";
		if($count >1) {$authorList[$count-1] = "and ".$authorList[$count-1];}
		
		$authors = join(", ", @authorList);
		my $i2 = $i + 1;
		print OUTFILE "<TR>\n\t<TD>\t[$i2]\n";
		#print $authors; #in the HTM file
		print OUTFILE "\t<TD>".$authors."\. ";
		}

	# TITLE GENERATION
	my $title = $entry{"title"};
	if(!exists($entry{"title"}))
		{ print "Warning: empty Title in Article\n"; }
	else
		{ print OUTFILE " $title. ";}
	
	# JOURNAL#
	if(!exists($entry{"journal"}))
		{ print "Warning: empty Journal in Article\n";}
	else
		{
		my $journal = $entry{"journal"};
		print OUTFILE "  <I>$journal, <\/I>";
		}
	

	# Volume, Number, Pages, Month and year 	
	if(exists($entry{"volume"}))
		{ my $volume = $entry{"volume"}; print OUTFILE $volume." ";}
	if(exists($entry{"number"}))
		{ my $number = $entry{"number"}; print OUTFILE "($number) "; }
	
	if(exists($entry{"pages"}))
		{
		if ( exists($entry{"volume"}) || exists($entry{"number"} ) )
			{ my $pages = $entry{"pages"}; print OUTFILE "\: $pages"; }
		else
			{ my $pages = $entry{"pages"}; print OUTFILE " pages $pages"; }
		}



	if(exists($entry{"month"}))
		{
		my $monthVar = $month{$entry{"month"}}; 
		print OUTFILE ", $monthVar ";
		}
	if(exists($entry{"year"}))
		{
		my $year = $entry{"year"};
		print OUTFILE " ".$year."\.\n";
		}
	 else	
		{print "Warning: Year missing in Article\n"; }
	print OUTFILE "<\/TR>\n";
	}
	

	# BOOK GENERATION ################################################
	# REQUIRED - AUTHOR/ EDITOR, TITLE, PUBLISHER, YEAR
	# OPTIONAL - VOL, NO, SERIES, ADDR, EDITION, MONTH
	##################################################################

	elsif(lc($entry{"type"}) eq lc("book"))
        {
                                                                                                                             
        # remove the AND and replace them with commas to seperate all authors and last one by "and"
        # AUTHOR ADDITION 
        if(!exists( $entry{"author"}) )
                { print "Warning: empty Author in Book\n";}
        else
                {
		
                my $authors = $entry{"author"};
		
                my @authorList = split(/ and /, $authors);
                my $count = @authorList."\n";
                if($count >1) {$authorList[$count-1] = "and ".$authorList[$count-1];}
                                                                                                                             
                $authors = join(", ", @authorList);
		my $i2 = $i + 1;
                print OUTFILE "<TR>\n\t<TD>\t[$i2]\n";
                #print $authors; #in the HTM file
                print OUTFILE "\t<TD>".$authors."\. ";
                }
                                                                                                                             
        # TITLE GENERATION 
        my $title = $entry{"title"};
        if(!exists($entry{"title"}))
     		{ print "Warning: empty Title in Book\n"; }
        else
                { print OUTFILE "<I> $title.<\/I> ";}
                                                                                                                             
        # NUMBER#
        if(exists($entry{"number"}))
              {} 
        else
                {
                my $number = $entry{"number"};
                print OUTFILE "  Number $number in";
                }
                                                                                                                             
                                                                                                                             
        # Series, Pages, Editors(mandatory), Publisher, address, volume, edition, Month and year 
        if(exists($entry{"series"}))
                { my $series = $entry{"series"}; print OUTFILE "($series)"; }
	
        if(!exists( $entry{"editor"}) )
                { print "Warning: empty Editor in Book\n";}
        else
                {
		
                my $editors = $entry{"editor"};
		
                my @editorList = split(/ and /, $editors);
                my $count = @editorList."\n";
                if($count >1) {$editorList[$count-1] = "and ".$editorList[$count-1];}
                                                                                                                             
                $editors = join(", ", @editorList);
               
                print OUTFILE $editors."\, editors\. ";
                }                       
                                                
	if(exists($entry{"publisher"}))
                { my $publisher = $entry{"publisher"}; print OUTFILE "($publisher)"; }
	else
		{ print "Warning: Empty Publisher in Book\n"; }
                                                                                 
	if(exists($entry{"address"}))
                { my $address = $entry{"address"}; print OUTFILE "($address)"; }
                                                                                     
    
	if(exists($entry{"volume"}))
		  { my $volume = $entry{"volume"}; print OUTFILE $volume.", ";}

        if(exists($entry{"edition"}))
		  { my $edition = $entry{"edition"}; print OUTFILE $edition." Edition, ";}
                              
      
                                                                                                                             
                                                                                                                             
                                                                                                                             
        if(exists($entry{"month"}))
                {
                my $monthVar = $month{$entry{"month"}};
                print OUTFILE ", $monthVar ";
                }
        if(exists($entry{"year"}))
                {
                my $year = $entry{"year"};
                print OUTFILE " ".$year."\.\n";
                }
         else   
              {print "Warning: Year missing in Book\n";}
        print OUTFILE "<\/TR>\n";
        }


	# INPROCEEDINGS #####################################################
	# REQUIRED - AUTHOR, TITLE, BOOKTITLE, YEAR
	# OTPIONAL - EDITOR, VOL, NO, SERIES , PAGES, MONTH, ORGANSATION, PUBLISHR, ADDR
	#####################################################################

 	elsif(lc($entry{"type"}) eq "inproceedings" )
        {
                                                                                                                             
        # remove the AND and replace them with commas to seperate all authors and last one by "and"
        # AUTHOR ADDITION 
        if(!exists( $entry{"author"}) )
                { print "Warning: empty Author in Inproceedings\n";}
        else
                {
		
                my $authors = $entry{"author"};
		
                my @authorList = split(/ and /, $authors);
                my $count = @authorList."\n";
                if($count >1) {$authorList[$count-1] = "and ".$authorList[$count-1];}
                                                                                                                             
                $authors = join(", ", @authorList);
		my $i2 = $i + 1;
                print OUTFILE "<TR>\n\t<TD>\t[$i2]\n";
               
                print OUTFILE "\t<TD>".$authors."\. ";
                }
                                                                                                                             
        # TITLE GENERATION 
        my $title = $entry{"title"};
        if(!exists($entry{"title"}))
     		{ print "Warning: empty Title in Inproceedings\n"; }
        else
                { print OUTFILE " $title. ";}
                                                                                                                             
                                                                                                                      
                                                                                                                             
        # Editor, booktitlem, volume Month and year        
        if(!exists( $entry{"editor"}) )
                { }
        else
                {
		
                my $editors = $entry{"editor"};
		
                my @editorList = split(/ and /, $editors);
                my $count = @editorList."\n";
                if($count >1) {$editorList[$count-1] = "and ".$editorList[$count-1];}
                                                                                                                             
                $editors = join(", ", @editorList);
                
                print OUTFILE $editors."\, editors\. ";
                }                       
                
	my $bookTitle = $entry{"booktitle"};
        if(!exists($entry{"booktitle"}))
     		{ print "Warning: empty Book Title in Inproceedings\n"; }
        else
                { print OUTFILE " <I>$bookTitle. ,<\/I>";}
                                                           

        
	if(exists($entry{"publisher"}))
                { my $publisher = $entry{"publisher"}; print OUTFILE "($publisher)"; }
	else
		{ print "Warning: Empty Publisher in Inproceedings\n"; }
                                                                                 
	if(exists($entry{"address"}))
                { my $address = $entry{"address"}; print OUTFILE "$address"; }
         
	  if(exists($entry{"organization"}))
                { my $organization = $entry{"organization"}; print OUTFILE " $organization "; }                                                                          
    
	if(exists($entry{"volume"}))
		  { my $volume = $entry{"volume"}; print OUTFILE $volume.", ";}

        if(exists($entry{"number"}))
		  { my $number = $entry{"number"}; print OUTFILE "($number)\n";}
                              
      if(exists($entry{"series"}))
                { my $series = $entry{"series"}; print OUTFILE " in $series"; }
	           
             if(exists($entry{"pages"}))
		{
		if ( exists($entry{"volume"}) || exists($entry{"number"} ) )
			{ my $pages = $entry{"pages"}; print OUTFILE "\: $pages"; }
		else
			{ my $pages = $entry{"pages"}; print OUTFILE " pages $pages"; }
		}


                                                                                                                
                                                                                                                             
                                                                                                                             
        if(exists($entry{"month"}))
                {
                my $monthVar = $month{$entry{"month"}};
                print OUTFILE ", $monthVar ";
                }
        if(exists($entry{"year"}))
                {
                my $year = $entry{"year"};
                print OUTFILE " ".$year."\.\n";
                }
         else   
              {print "Warning: Year missing in Inproceedings\n";}
        print OUTFILE "<\/TR>\n";
        }


	


	# SINCE IT IS NOT POSSIBLE TO PARSE AND DISPLAY ALL SUCH types of documents, we define a generic type whic displays all the fields provided on the bibtex entry
	# eg if we provide PhD thesis with author, title, school and year, it will parse this and print it correctly in thecorrect format.
	# however if the PhD thesis also includes organisation or booktitle, which are not supposed to be there, such output will aslo be displyd -this lack of error checking  is the limitattion of the system.
	# thus the onus of providing the correct bibtex formats for each type of document i.e. for each entry lies on the user
	#DEFAULT RESULT IS THAT ALL FIELDS VALID IN A BIBTEX FILE IRRESPECTIVE OF THE TYPE OF DOCUMENT IS DISPLAYED
	

	# no warnings are displayed in the generic definition
	else
	{
                                                                                                                             
        # remove the AND and replace them with commas to seperate all authors and last one by "and"
        # AUTHOR ADDITION 
        if(exists( $entry{"author"}) )
            
                {
		
                my $authors = $entry{"author"};
		
                my @authorList = split(/ and /, $authors);
                my $count = @authorList."\n";
                if($count >1) {$authorList[$count-1] = "and ".$authorList[$count-1];}
                                                                                                                             
                $authors = join(", ", @authorList);
		my $i2 = $i + 1;
                print OUTFILE "<TR>\n\t<TD>\t[$i2]\n";
    
                print OUTFILE "\t<TD>".$authors."\. ";
                }
                                                                                                                             
        # TITLE GENERATION 
        my $title = $entry{"title"};
        if(exists($entry{"title"}))
     		{ print OUTFILE " <I>$title. <\/I>";}
                                                                                                                             
         # JOURNAL#########
        if(exists($entry{"journal"}))
                {
                my $journal = $entry{"journal"};
                print OUTFILE "  <I>$journal, <\/I>";
                }
                                                                                                                             
                                                                                                              
                                                                                                                             
        # Editor, booktitlem, volume Month and year and all other such data
        
        if(exists( $entry{"editor"}) )
               {
		
                my $editors = $entry{"editor"};
		
                my @editorList = split(/ and /, $editors);
                my $count = @editorList."\n";
                if($count >1) {$editorList[$count-1] = "and ".$editorList[$count-1];}
                                                                                                                             
                $editors = join(", ", @editorList);
               
                print OUTFILE $editors."\, editors\. ";
                }                       
                
	my $bookTitle = $entry{"booktitle"};
        if(exists($entry{"booktitle"}))
     
                { print OUTFILE " <I>$bookTitle. <\/I>";}
                                                           

        
	if(exists($entry{"publisher"}))
                { my $publisher = $entry{"publisher"}; print OUTFILE "($publisher)"; }
	
                                                                                 
	if(exists($entry{"address"}))
                { my $address = $entry{"address"}; print OUTFILE "$address"; }
         
	  if(exists($entry{"organization"}))
                { my $organization = $entry{"organization"}; print OUTFILE " $organization "; }                                                                          
    
	if(exists($entry{"volume"}))
		  { my $volume = $entry{"volume"}; print OUTFILE $volume.", ";}

        if(exists($entry{"number"}))
		  { my $number = $entry{"number"}; print OUTFILE "($number)\n";}
                              
        if(exists($entry{"series"}))
                { my $series = $entry{"series"}; print OUTFILE " in $series"; }
	           
        if(exists($entry{"pages"}))
		{
		if ( exists($entry{"volume"}) || exists($entry{"number"} ) )
			{ my $pages = $entry{"pages"}; print OUTFILE "\: $pages, "; }
		else
			{ my $pages = $entry{"pages"}; print OUTFILE " pages $pages, "; }
		}


                                                                                                                
                                                                                                                             
                                                                                                                             
        if(exists($entry{"month"}))
                {
                my $monthVar = $month{$entry{"month"}};
                print OUTFILE ", $monthVar ";
                }
        if(exists($entry{"year"}))
                {
                my $year = $entry{"year"};
                print OUTFILE " ".$year."\.\n";
                }
        
        print OUTFILE "<\/TR>\n";
        }


	
	}	# end of for loopFILE "<\/TABLE>";
}




