#!/usr/bin/perl  
use strict;


# read the name of the squidlog file from the command line into fileName
my $fileName = $ARGV[0];

# open the squid log file and test whether it can be opened and whether read permission is given
unless(open(SQUID_FILE, $fileName))
{
	if(!(-e $fileName))
		{
 		die("Input File does not exist");
		}
	elsif(!(-r $fileName))
		{
		die("You are not allowed to read the input file");
		}
	else
		{
		die ("File does not exist");
		}
}

# open the output file 
open ( OUTFILE, ">$ARGV[1]" ) || die("Output File could not be found ");


# the various parameters stored by the module
my %codeArray;
my %associateArray;
my %topUsers;
my %dataArray;
my %timeArray;
my $firstTimeStamp;
my $timeStampCount = 0;
my $tcpHits = 0;
my $tcpMisses = 0;
my $errors = 0;


# call the modules
&Begin;			# beginning welcome message
 
&ReadSplitInputFile; 	# extract IP adresses and various parameters into the respective hash arrays and scaler variables

print OUTFILE "\tNUMBER OF REQUESTS BY EACH IP\n\n";
&NoOfRequests;		# find no of requests by each IP addr
&Line;

print OUTFILE "\tTOP 5 USERS WITH HIGHEST NUMBER OF REQUESTS\n\n";
&Top5Requests;		# find top 5 users with highest no of requests
&Line;

print OUTFILE "\tTHE NO OF UNIQUE DOMAINS & THE MOST FREQUENTLY ACCESSED DOMAINS BY THE TOP 5 REQUESTOR IPS\n\n";
&UniqueDomains;         # find unique domains of the top 5 ip
&Line;

print OUTFILE "\tAMOUNT OF DATA DOWNLOADED BY EACH IP & TOP 5 IPS THAT DOWNLOADED LARGEST AMOUNT OF DATA\n\n";
&MaxData;               # find users with max data downloaded
&Line;

print OUTFILE "\tRATIO OF CACHE HITS TO CACHE MISSES\n\n";
&CacheRatio;            # find cache hit to miss ratio
&Line;

print OUTFILE "\tHIT RESPONSE CODE ANALYSIS\n\n";
&HitResponseCodeAnalysis;   # find the Hit response codes and their counts
&Line;

print OUTFILE "\tTOTAL NO OF TIMESTAMPS PER 5 MIN, MAXIMUM AND AVERAGE\n\n";
&TimeAnalysis;		# find the no of hits in every 5 minute interval
&Line;

&End;			# end formatting and messages


#######################################################
# END OF MAIN - DEFINE THE SUBROUTINES..........
#######################################################

# welcome screen
sub Begin
{
for(my $i = 0; $i <= 80; $i++)
	{print OUTFILE "*";}
print OUTFILE "\n\n\t\t WELCOME TO THE SQUID LOG ANALYSER\n\n";
for(my $i = 0; $i < 80; $i++)
	{print  OUTFILE "*";}
print  OUTFILE "\n\n";
}	# end of BEGIN


# on exit display
sub End
{

print OUTFILE "This file was generated as output from squid.pl written as part of our CS701 Perl Assignment 2\n";
print OUTFILE "For any suggestions or criticism, please mail us.\n";
print OUTFILE "The group is G6:\n Kriti Puniyani\t04305012\tkriti\@cse.iitb.ac.in\n Gautham Anil\t04305013\tgautham_anil\@cse.iitb.ac.in\n";
print STDERR "Done Processing - the output has been stored in file : $ARGV[1]\n\n";
}	# end of END


# display a line of 80 - for demarcation purposes
sub Line
{
print  OUTFILE "\n\n";
for(my $i = 0; $i <= 80; $i++)
	{print  OUTFILE "-";}
print  OUTFILE "\n\n";
}	# end of line


# subroutine to compare 2 ip addresses
sub by_IP
{
my @input1 = split(/\./, $a);
my @input2 = split(/\./, $b);
for(my $i = 0; $i <=3; $i++)
	{
	if($input1[$i] > $input2[$i])
		{ return 1; }
	elsif($input1[$i] < $input2[$i])
		{ return -1; }
	}
return 0;
}	# end of by_IP


# read the input file and extract IPaddr associateArray from it
sub ReadSplitInputFile
{
    
my $line = <SQUID_FILE>;

# find first timestamp
$firstTimeStamp = (split (/\./, $line))[0];

my $interval = 1;

# read the input file line by line
while($line ne "")
	{
	my @lineSplit = split (/[\t ]+/, $line);
	
	# find lineSplit[2] i.e.IP address and increment its count by 1
	# if it doesnt already exist, then its count = 0
	if(!exists($associateArray{$lineSplit[2]}) )
		{
		$associateArray{$lineSplit[2]} = 1;
		}
	else
		{
		$associateArray{$lineSplit[2]} += 1;
		}


	#find the datasize per IP addr 
	if(!exists($dataArray{$lineSplit[2]}) )
		{
		$dataArray{$lineSplit[2]} = $lineSplit[4];
		}
	else
		{
		$dataArray{$lineSplit[2]} += $lineSplit[4];
		}


	# find the no of TCP hits and misses
	if( $lineSplit[3] =~ /^TCP_MISS/)
	        {$tcpMisses++;}
	elsif( $lineSplit[3] =~ /^TCP_HIT/)
	        {$tcpHits++;}
	else 
         	{$errors++;}
	

        # find the no of different codes returned and their values
        
        my @codeSplit = split ('\/', $lineSplit[3]);

        if( $codeSplit[1] =~ /[0-9]*$/)
		{
		if(!exists($codeArray {$codeSplit[1]} ))
			{
			$codeArray{$codeSplit[1]} = 1;
			}
		else
			{
			 $codeArray{$codeSplit[1]} += 1;
			}	
		}
	else
		{    
		print  STDERR "Error : Code response does not seem to be accurate\n";
		}
                                                                                                                      

	# create timestamp intervals of 300 sec each and count the no of timestamps occuring in that interval
	my $integerValue = ( split(/\./, $lineSplit[0]) )[0] - $firstTimeStamp - 300*($interval - 1);

	if($integerValue >= 300)
		{		
		$interval += 1;
		$timeArray{$interval} = 1;
		}
	else
		{
		$timeArray{$interval} += 1;
		}	
	$timeStampCount++;

	$line = <SQUID_FILE>;

	}	# end of while loop
close(SQUID_FILE);
}	# end of ReadSplitInputFile


# find the no of requests issued by each IP addr and print in sorted order
sub NoOfRequests
{
my $key;
my $value;

print OUTFILE  "The no of requests from each IP is: (in sorted order of IP)\n\n";
print  OUTFILE "IP Address\tNo. of Requests\n";
foreach $key (sort by_IP keys(%associateArray))
	{
	print  OUTFILE ("$key\t$associateArray{$key}\n");
	}


}            # end of NoOfRequests


# find top 5 users which issue max no of requests and sort them by IP addr
sub Top5Requests
{

my $key;
my $value;
my $count = 1;		# count and print only top 5 results

print  OUTFILE "The top 5 users with the highest no. of requests are (in sorted order of IP):\n\n";
print  OUTFILE "IP Address\tNo. of Requests\n";
# sort the array based on value and not key and in descending order

foreach $key (sort { $associateArray{$b} <=> $associateArray{$a} } keys(%associateArray))
	{
	$topUsers{$key} = $associateArray{$key};
	if($count++ >= 5)
		{last;}
	}
$count = 1;
foreach $key (sort by_IP keys(%topUsers))
        {
        print OUTFILE ("$key\t$topUsers{$key}\n");
        if($count++ >= 5)
                {last;}
        }



}             # end of Top5Requests




# count the no of unique domains and the domain most frequently accessed by the top 5 users
sub UniqueDomains
{
# for each topUser do
# find no of domains and store in domainArray with key = domain & value = count
# agian for each user do in for display @domainArray(no of domains) and max on value

my $ip;
my $domain = 0;
print STDERR "Please be patient.....processing your input\n\n";
foreach $ip ( sort by_IP keys(%topUsers))
	{
	my %domainArray;
	open(SQUID_FILE, $fileName);
	my $line = <SQUID_FILE>;
	
	# read the input file line by line
	while($line ne "")
		{
		my @lineSplit = split (/[\t ]+/, $line);
		
		# find lineSplit[2] i.e.IP address and check if it is $ip
		# if yes then count++ store in domainArray
		if($lineSplit[2] eq $ip)
			{
			$domain = (split(/\/+/, $lineSplit[6]))[1];
			if(!exists( $domainArray{$domain} ) )
				{
				$domainArray{$domain} = 1;
				}
			else
				{
				$domainArray{$domain} += 1;
				}
			}    # end of if ip addr is found
		$line = <SQUID_FILE>;
		}    # end of while loop

	# scan the domainArray and find the element that occurs max times
	
	print  OUTFILE "\n\nFor requestor $ip : \n";
	my $uniqueCount = 0;
	
	# count the no of elements in the array
	my $value;
	while(($domain, $value) = each(%domainArray))
		{ $uniqueCount++; }
	
	print OUTFILE ( "The Number of Unique Domain Names are : \t\t".$uniqueCount."\n");
	
	# find max element
	my $max = 0;
	foreach $domain (%domainArray)
		{
		if( $domainArray{$max} < $domainArray{$domain} )	
			{$max = $domain;}
		}
	print  OUTFILE "The domain which is accessed maximum no of times : \t$max\n";
	close (SQUID_FILE);
	
	# ensure that domainArray is NULL again
	my %c;
	%domainArray = %c;
	$domain = 0;

	}	# end of for each ip

}		# end of UniqueDomains


# calculate the amount of data downloaded by each ip and top 5 inputs that download max amount of data
sub MaxData
{
my %top5Users;
my $line;

#sort the input file according to data size and then display top 5 sorted according to IP addr 

print  OUTFILE "The amount of data downloaded by each IP is: (in sorted order of IP)\n\n";
print  OUTFILE "IP Address\tAmount of Data Downloaded\n";

my $key;
my $value;

foreach $key (sort by_IP keys(%dataArray))
	{
	print OUTFILE ("$key\t$dataArray{$key}\n");
	}

&Line;

print  OUTFILE  "The top 5 users with the highest no. of data downloaded are (in sorted order of IP):\n\n";
print  OUTFILE "IP Address\tAmount of Data Downloaded\n";
my $count = 1;		# count and print only top 5 results


# sort the array based on value and not key and in descending order

foreach $key (sort { $dataArray{$b} <=> $dataArray{$a} } keys(%dataArray))
	{
	$top5Users{$key} = $dataArray{$key};
	if($count++ >= 5)
		{last;}
	}
$count = 1;
foreach $key (sort by_IP keys(%top5Users))
        {
        print OUTFILE ("$key\t$top5Users{$key}\n");
        if($count++ >= 5)
                {last;}
        }


}                         # end of Max Data



# find ratio of Cache Misses to cache hits
sub CacheRatio
{
print  OUTFILE "No of TCP HITS is  \t\t\t : $tcpHits\n";
print  OUTFILE "No of TCP MISSES is\t\t\t : $tcpMisses\n";
print  OUTFILE "No of TCP Messages neither MISSES nor HITS (TCP_DENIED, TCP_CLIENT_REFRESH_MISS etc.) are \t : $errors\n";
my $ratio = $tcpHits/$tcpMisses;
print  OUTFILE "\nRatio of Cache Hits to Cache Misses in the log is\t\t : $ratio\n";
}                     # end of CacheRatio


sub HitResponseCodeAnalysis
{
my $key;

# define a hash array that hard code the relationship between a code and its message 
# eg.  code 200 is represented as "code 200 - OK"
my %codeRelation = (
000, "Undefined response code - Code", 
"200", "Code 200 - OK                     ", 
"202", "Code 202 - Accepted Code          ", 
"204", "Code 204 - No Content Code        ",
"205", "205 - Reset Content Code          ", 
"206", "206 - Partial Content Code        ",
"300", "300 - Multiple Choices Code       ",
"301", "301 - Moved Permanently Code      ", 
"302", "302 - Found Code                  ",
"303", "303 - See Other Code              ",
"304", "304 - Not Modified Code           ",
"307", "307 - Moved Temporarily Code      ", 
"400", "400 - Bad Request Code            ",
"401", "401 - Unauthorized Code           ",
"402", "402 - Payment Required            ",
"403", "Code 403 - Forbidden              ",
"404", "Code 404 - Not Found              ",
"405", "Code 405 - Method Not Allowed     ",
"406", "Code 406 - Not Acceptable         ",
"407", "407 - Proxy Authentication Required",
"408", "Code 408 - Request Timeout        ",
"409", "Code 409 - Conflict               ",
"410", "Code 410 - Gone                   ",
"411", "Code 411 - Length Required        ",
"412", "Code 412 - Precondition Failed    ",
"413", "Code 413 - Request Entity Too Large",
"414", "Code 414 - Request-URI Too Long   ",
"415", "Code 415 - Unsupported Media Type ",
"416", "416 - Requested Range Not Satisfiable",
"500", "Code 500 - Internal Server Error  ",
"501", "Code 501 - Not Implemented        ",
"502", "Code 502 - Bad Gateway            ",
"503", "Code 503 - Service Unavailable    ",
"504", "Code 504 - Gateway Timeout        ");


print  OUTFILE "Hit By Response Code Analysis: (In sorted order of the response code in increasing order)\n\n";
print  OUTFILE "Response Code\t\t\t\tCount\n\n";

foreach $key (sort keys(%codeArray))
	{
	if( exists($codeRelation{$key} ) ) 
		{ print  OUTFILE ("$codeRelation{$key}\t$codeArray{$key}\n");}
	else
		{ print  OUTFILE ("$codeRelation{000} $key\t$codeArray{$key}\n"); }
        }


}                         # end of HitResponseCodeAnalysis


# Find total number of hits per 5 mins. Give Maximum number of hits and Average number of hits per 5 mins. 
sub TimeAnalysis
{
my $key;
my $max = 0;
my $intervalCount = 0;
print  OUTFILE "Time Analysis of the Log:\n\nTotal No of hits per 5 minute interval:\n";
print  OUTFILE "Time Group\tStarting time\tEnding Time\tNo. of Hits:\n\n";	

foreach $key (sort { $a <=> $b } keys(%timeArray))
	{
	print OUTFILE ("$key\t\t".($firstTimeStamp+(($key-1)*300))."\t".($firstTimeStamp+$key*300)."\t".$timeArray{$key}."\n");
	if( $timeArray{$key} > $timeArray{$max} )
		{ $max = $key; }
	$intervalCount++;
	}

print  OUTFILE "\nTotal no of timestamps (inputs) analysed = $timeStampCount\n";

print  OUTFILE "Total no of intervals = ".$intervalCount."\n";
print  OUTFILE "\nAverage no of timestamps in any interval = ".($timeStampCount/$intervalCount)."\n";

print  OUTFILE "\nMaximum no of hits in any interval is $timeArray{$max} for interval $max \nwith starting timestamp ".($firstTimeStamp+(($max-1)*300))." and ending timestamp ".($firstTimeStamp+$max*300)."\n\n";
}		# end of TimeAnalysis



